#!/usr/bin/env python3
import sys
import os
import random
import string
import pexpect
import pexpect.pxssh as pxssh

if not os.path.exists('./var'): os.makedirs('./var')
import logging
LOG_PATH = './var/log.log'
#logging.basicConfig(filename=LOG_PATH,level=logging.INFO)
logging.basicConfig(filename=LOG_PATH,level=logging.DEBUG)

import configparser
CONFIG_PATH = 'data/config.ini'
config = configparser.ConfigParser()
config.read(CONFIG_PATH)

# Display all logging in STDOUT:
root = logging.getLogger()
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
root.addHandler(ch)
class ExceptionSCP(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

def scp(password, command):
    """File copy over SSH, using the standard scp tool"""
    logging.info('Copying with SCP: scp {}'.format(command))
    child = pexpect.spawn('scp ' + command)
    try:
        # Borrowed some ideas from
        # http://pexpect.readthedocs.org/en/latest/_modules/pexpect/pxssh.html
        #TODO: add a check for b'stdin: is not a tty\r\n' (You see it when ALREADY CONNECTED, but for some reason this message is only for root connections?)
        i = child.expect([pexpect.EOF, "(?i)are you sure you want to continue connecting", "(?i)(?:password)|(?:passphrase for key)", "(?i)permission denied", "(?i)terminal type", "(?i)connection closed by remote host", "(?i)No route to host"], timeout=120)
        logging.debug('index == ' + str(i))
        if i==0: # EOF
            logging.debug('Have the key, everything is fine')
        elif i==1: # are you sure you want to continue connecting
            logging.debug("Connecting for the first time, saying Yes")
            child.sendline("Yes\n")
            j = child.expect([pexpect.EOF, "(?i)(?:password)|(?:passphrase for key)"])
            if j==0:
                logging.debug('Have the key, everything is fine')
            elif j==1:
                logging.debug("Entering password")
                child.sendline(password + '\n')
                child.expect(pexpect.EOF, timeout=5)
            else:
                error_text = "CAN'T CONNECT (unexpected error): " + 'i = ' + str(i) + ' ' + str(child.before)
                logging.warning(error_text)
                raise ExceptionSCP(error_text)
        elif i==2: # password or passphrase
            logging.debug("Entering password")
            child.sendline(password + '\n')
            child.expect(pexpect.EOF, timeout=5)
        elif i==6: # No route to host
            error_text = "CAN'T CONNECT (NO ROUTE TO HOST): " + str(child.before)
            logging.warning(error_text)
            raise ExceptionSCP(error_text)
        else:
            error_text = "CAN'T CONNECT (unexpected error): " + 'i = ' + str(i) + ' ' + str(child.before)
            logging.warning(error_text)
            raise ExceptionSCP(error_text)
    except pexpect.TIMEOUT:
        error_text = "CAN'T CONNECT (TIMEOUT): " + str(child.before)
        logging.warning(error_text)
        if child.isalive():
            logging.warning("SSH process is still running. Killing it")
            child.kill(pexpect.signal.SIGINT) # Sending Ctrl+C
        if child.isalive():
            logging.error(child.isalive())
            logging.error("SPOOKY UNDEAD CHILD. Why is SSH process still running? Terminating")
            child.terminate()
            if child.isalive(): logging.error("IT'S STILL NOT DEAD")
        raise ExceptionSCP(error_text)

class ExceptionServerRun(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class ExceptionServerAptGet(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class Server:
    def __init__(self, ip, name, password, timeout=180):
        """Returns False if can't connect"""
        self._connected = False
        self._timeout = timeout # timeout for both connecting and when waiting for prompt
        self._s= pxssh.pxssh(timeout)
        self.name = name
        self.ip = ip
        self.password = ip
        try:
            self._s.login(ip, name, password)
            self._connected = True
        except pxssh.ExceptionPxssh as inst: logging.warning("Can't connect to {}@{}, details: \n{}".format(name, ip, inst))
        except pexpect.EOF as inst: logging.warning("Can't connect to {}@{}, details: \n{}".format(name, ip, inst))
    def __del__(self):
        self.logout()

    def logout(self):
        if self._connected:
            logging.info('Logging out of {}@{}'.format(self.name,self.ip))
            try:
                self._s.logout()
                self._connected = False
            except pexpect.TIMEOUT as inst:
                logging.error("Can't logout from {}@{}, PROBABLY FROZEN COMMAND. KILLING THE PROCESS. details: \n{}".format(name, ip, inst))
                self._s.close(force=True)

    def is_connected(self):
        if self._connected: return True
        else: return False

    def run(self, command):
        logging.debug('Running command on {}@{}: {}'.format(self.name, self.ip, command))
        try:
            self._s.sendline(command)
            if not self._s.prompt(self._timeout):
                error_text = 'THE COMMAND IS STILL RUNNING:{} :\n{}'.format(command, self._s.read())
                logging.warning(error_text)
                raise ExceptionServerRun(error_text)
            output = self._s.before
            logging.debug('OUTPUT: {}'.format(output))
            return output
        except pexpect.TIMEOUT as e:
            error_text = "Timeout when sending command {} (or waiting for prompt), details:\n{}".format(command, e)
            raise ExceptionServerRun(error_text)

    def apt_get(self, apt_command, *packages):
        """apt_command can be install, remove, purge and other"""
        logging.debug('Running apt-get: {} {}'.format(apt_command, ' '.join(packages)))
        try:
            output = self.run('apt-get -yq ' + apt_command + ' ' + ' '.join(packages))
            if b'is already the newest version' in output:
                logging.warning("One of the packages in '{}' was already installed on the server:\n{}".format(' '.join(packages), output))
            if b'Could not get lock' in output or \
               b'Unable to locate package' in output or b'Unable to lock directory' in output:
                raise ExceptionServerAptGet('Errors found when running apt-get {} for the packages : {}, details:\n{}'.format(apt_command, ' '.join(packages), output))
        except ExceptionServerRun as e:
            raise ExceptionServerAptGet('Errors found when running apt-get {} for the packages: {}, details:\n{}'.format(apt_command, ' '.join(packages), e))
        return output

    def generate_exim_ssl_key(self, domain):
        logging.debug('Generating SSL key for Exim4 (MIGHT BREAK Pexpect), domain: {}'.format(domain))
        try:
            #self._s.sendline('openssl req -newkey rsa:2048 -keyout exim.key -out exim.csr -days 3650 -nodes')
            #self._s.sendline('/usr/bin/openssl req -x509 -newkey rsa:2048 -keyout /etc/exim4/ssl/exim.key -out /etc/exim4/ssl/exim.crt -days 9000 -nodes')
            self._s.sendline('/usr/share/doc/exim4-base/examples/exim-gencert')
            self._s.sendline('\n\n\n\n')
            self._s.sendline(domain + '\n') # ADDING ANOTHER '\n' here will break Pexpect
            self._s.prompt(self._timeout)
            self.run('mv /etc/exim4/exim.key /etc/exim4/exim.crt /etc/exim4/ssl/')
            #self.run('openssl req -new -key /etc/exim4/ssl/exim.key -out /etc/exim4/ssl/exim.csr')
        except pexpect.TIMEOUT as e:
            error_text = "Timeout when generating an SSL certificate for {}, details: \n{}".format(domain, e)
            raise ExceptionServerRun(error_text)

with open('data/email-names.txt') as f: mailbox_list = f.readlines()
mailboxes = iter(mailbox_list)

result = open('var/email-total.txt', 'w')
result_AMS = open('var/AMS.txt', 'w')

class ExceptionServerConfiguration(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

def configureServer(ip, name, password, domain, dns_zone_template):
    # Remove SSH key from ~/.ssh/known_hosts (in case the server was reinstalled)
    child = pexpect.spawn('ssh-keygen -R ' + ip)
    child.expect(['(?i)known_hosts updated', '(?i)No such file or directory'], timeout=2)

    s = Server(ip, name, password, timeout=180)
    if not s.is_connected():
        error_text = "Can't login into {}@{}".format(name, ip)
        logging.error(error_text)
        raise ExceptionServerConfiguration(error_text)

    if b'Debian GNU/Linux 7' not in s.run('cat /etc/issue.net'):
        raise ExceptionServerConfiguration('Not a Debian 7 system')

    if b'en_US.UTF-8' not in s.run('env | grep LANG'):
        logging.info("LANG is not en_US.UTF-8, trying to change")
        s.run("echo 'en_US.UTF-8 UTF-8' >> /etc/locale.gen")
        s.run("echo 'LANG=en_US.UTF-8' > /etc/default/locale")
        s.run('/usr/sbin/locale-gen')
        s.run('export LANG=en_US.UTF-8')
        s.run('export LC_ALL=en_US.UTF-8')
        # Now test
        if b'No such file or directory' not in s.run('ls nonexistentfileforsure'):
            raise ExceptionServerConfiguration("The server ain't speaking no English")

    if b'No such file or directory' not in s.run('ls /usr/local/mgr5'):
        raise ExceptionServerConfiguration("Can't configure servers with ISPmanager installed")

    s.run("echo '{}' > /etc/mailname".format(domain))

    # Change /etc/hostname and /etc/hosts
    if b'No such file or directory' in s.run('ls /etc/hostname.backup'):
        s.run('cp /etc/hostname /etc/hostname.backup')
    s.run("echo '{}' > /etc/hostname".format(domain))
    s.run('/etc/init.d/hostname.sh start')
    if b'No such file or directory' in s.run('ls /etc/hosts.backup'):
        s.run('cp /etc/hosts /etc/hosts.backup')
    s.run("""echo '127.0.0.1 localhost.localdomain localhost
{ip} {domain} {domainmask}
::1     localhost ip6-localhost ip6-loopback
' > /etc/hosts """.format(ip=ip, domain=domain, domainmask=domain.split('.')[0]))

    if b'No such file or directory' in s.run('ls /etc/apt/sources.list.backup'):
        s.run('cp /etc/apt/sources.list /etc/apt/sources.list.backup')
    s.run("""echo 'deb http://ftp.de.debian.org/debian wheezy main contrib non-free
deb http://security.debian.org/ wheezy/updates main contrib non-free
deb http://ftp.de.debian.org/debian wheezy-updates main' > /etc/apt/sources.list""")
    s.apt_get('update', '')
    s.run('DEBIAN_FRONTEND=noninteractive')

    # Remove all old configurations before installing packages
    s.run('/etc/init.d/sendmail stop')
    s.apt_get('remove', 'sendmail-base', 'sendmail-bin', 'sendmail')
    s.apt_get('purge', 'exim4-base', 'exim4-daemon-heavy', 'bind9', 'opendkim-tools')
    s.run('rm -rvf /etc/bind/*')
    s.run('rm -rvf /etc/exim4/ssl')
    # Remove all human Linux users (potentionally dangerous for production servers)
    #command = """for user in `awk -F: '$3 >= 1000 && $1 != "nobody" {print $1}' /etc/passwd` ;do userdel $user;done"""
    #s.run(command)

    s.apt_get('install', 'exim4', 'exim4-daemon-heavy', 'bind9', 'opendkim-tools', 'bsd-mailx')
    s.run('mkdir /etc/exim4/ssl')

    s.run("""sed -i "s/dc_eximconfig_configtype=.*/dc_eximconfig_configtype='internet'/" /etc/exim4/update-exim4.conf.conf""")
    s.run("""sed -i "s/dc_local_interfaces=.*/dc_local_interfaces=''/" /etc/exim4/update-exim4.conf.conf""")
    #s.run("""sed -i "s/dc_other_hostnames=.*/dc_other_hostnames='{}'/" /etc/exim4/update-exim4.conf.conf""".format(' '.join(subdomains_full)))
    s.run("""sed -i "s/dc_other_hostnames=.*/dc_other_hostnames='mail.{}'/" /etc/exim4/update-exim4.conf.conf""".format(domain))
    s.run('dpkg-reconfigure exim4-config -fnoninteractive')

    scp_command = 'data/exim4.conf.template {}@{}:/etc/exim4/'.format(name, ip)
    scp(password, scp_command)
    # Only needed after editing exim4.conf.template
    output = s.run('update-exim4.conf')
    if b'Invalid new configfile' in output: raise ExceptionServerConfiguration(output)
    output = s.run('/etc/init.d/exim4 restart')
    if b'Exim configuration error' in output or b'exim paniclog' in output:
        raise ExceptionServerConfiguration(output)

    # Create a user
    user_name = ''.join(random.SystemRandom().choice(string.ascii_lowercase + string.digits) for _ in range(6))
    user_password = ''.join(random.SystemRandom().choice(string.ascii_lowercase + string.digits) for _ in range(15))
    s.run('useradd -m -p $(openssl passwd -1 ' + user_password + ') ' + user_name)
    UID = s.run('id -u ' + user_name).decode('ascii').split('\r\n')[-2]
    GID = s.run('id -g ' + user_name).decode('ascii').split('\r\n')[-2]

    global result
    global result_AMS
    result.write(':'.join([ip, user_name, user_password]) + '\n')

    # Add a DNS zone line to /etc/bind/named.conf
    zoneline = 'zone "{domain}" {{ type master; file "/etc/bind/{domain}"; }};'.format(domain=domain)
    s.run("echo '{}' >> /etc/bind/named.conf".format(zoneline))

    global mailboxes # Access our global iterator
    exim4_aliases = ''
    exim4_passwd = ''
    exim4_domains = ''
    exim4_password_cram = ''

    dns_zone_file = dns_zone_template.format(domain=domain, ip=ip)

    subdomains = []
    for subdomain_count in range(0, int(config['Global']['sub_quantity'])):
        subdomain = ''.join(random.SystemRandom().choice(string.ascii_lowercase) for _ in range(int(config['Global']['subnames_Y'])))
        subdomains.append(subdomain)

    for subdomain in subdomains:
        subdomain_full = subdomain + '.' + domain

        # Add DNS records for subdomains
        subdomain_template = """
{subdomain} IN      A       {ip}
mail.{subdomain} IN      A       {ip}
{subdomain} IN      MX       10 mail
{subdomain} IN      MX       20 mail
pop.{subdomain}  IN      A       {ip}
smtp.{subdomain} IN      A       {ip}
{subdomain}      IN      TXT     "v=spf1 ip4:{ip} a mx ~all"
_dmarc.{subdomain}        IN      TXT     "v=DMARC1; p=none"
"""
        dns_zone_file += subdomain_template.format(ip=ip, subdomain=subdomain)

        # Generate DKIM keys
        command = '/usr/bin/opendkim-genkey -D /etc/exim4/ssl -d {} -s {} -r'.format(subdomain_full, subdomain_full)
        s.run(command)

    if not subdomains: names = [domain]
    else: names = [subdomain + '.' + domain for subdomain in subdomains]
    for subdomain_full in names:
        # Add subdomains to /etc/exim4/domains
        exim4_domains += '{subdomain}:{subdomain}::no\n'.format(subdomain=subdomain_full)

        for mailbox_count in range(0, int(config['Global']['emails'])):
            mailbox = next(mailboxes).strip()
            mailbox_full = mailbox + '@' + subdomain_full
            mailbox_home = '/home/' + user_name + '/email/' + subdomain_full + '/' + mailbox

            result.write(mailbox_full + '\n')
            result_AMS.write(ip + '|587|15|1|1|true|0|0|1|0|' + mailbox_full + '|' + config['Global']['email_pass'] + '|0|1|0|10|1||\n')

            # Generate /etc//etc/exim4/aliases
            # TODO: Maybe "send a copy" meant differently here. Check if forwarding works
            exim4_aliases += ':'.join([mailbox_full, config['Global']['email-copy'], 'user']) + '\n'

            # Generate /etc/exim4/passwords-cram for SMTP authorization
            exim4_password_cram += ':'.join([mailbox_full, config['Global']['email_pass']]) + '\n'

            # Generate /etc//etc/exim4/passwd. CURRENTLY NOT USED
            exim4_passwd += ':'.join([mailbox_full, UID, GID, '', mailbox_home, 'no']) + '\n'

    with open('var/temp_aliases', 'w') as f: f.write(exim4_aliases)
    with open('var/temp_passwd', 'w') as f: f.write(exim4_passwd)
    with open('var/temp_domains', 'w') as f: f.write(exim4_domains)
    with open('var/temp_passwords-cram', 'w') as f: f.write(exim4_password_cram)
    scp(password, 'var/temp_aliases {}@{}:/etc/exim4/aliases'.format(name,ip))
    scp(password, 'var/temp_passwd {}@{}:/etc/exim4/passwd'.format(name,ip))
    scp(password, 'var/temp_domains {}@{}:/etc/exim4/domains'.format(name,ip))
    scp(password, 'var/temp_passwords-cram {}@{}:/etc/exim4/passwords-cram'.format(name,ip))

    with open('var/temp_dns_zone_file', 'w') as temp_dns_zone_file:
        temp_dns_zone_file.write(dns_zone_file)
    # Copy the zone file with SCP
    scp_command = 'var/temp_dns_zone_file {}@{}:/etc/bind/{}'.format(name, ip, domain)
    scp(password, scp_command)

    # Append a DKIM key for the main domain
    command = '/usr/bin/opendkim-genkey -D /etc/exim4/ssl -d {} -s {} -r'.format(domain, domain)
    s.run(command)
    s.run('cat /etc/exim4/ssl/{}.txt >> /etc/bind/{}'.format(domain, domain))
    # Append a DKIM key for each subdomain
    for subdomain in subdomains:
        subdomain_full = subdomain + '.' + domain
        s.run(command)
        s.run('cat /etc/exim4/ssl/{}.txt >> /etc/bind/{}'.format(subdomain_full, domain))
        # Fix DKIM DNS record for each domain
        s.run("sed -i 's/{subdomain}.{domain}._domainkey/dkim._domainkey.{subdomain}/' /etc/bind/{domain}".format(subdomain=subdomain, domain=domain))
    # Fix DKIM DNS record for the main domain (ONLY SUBDOMAINS ARE USED FOR NOW, BUT JUST IN CASE)
    s.run("sed -i 's/{domain}._domainkey/dkim._domainkey/' /etc/bind/{domain}".format(domain=domain))

    s.generate_exim_ssl_key(domain) # THIS IS A KNOWN SOURCE OF BUGS, DO IT LAST
    s.run("chown -R Debian-exim:Debian-exim /etc/exim4/ssl /etc/exim4/passwd /etc/exim4/aliases /etc/exim4/domains /etc/exim4/passwords-cram")

    s.run('/etc/init.d/sendmail stop')
    output = s.run('/etc/init.d/exim4 restart')
    if b'Exim configuration error' in output or b'exim paniclog' in output:
        raise ExceptionServerConfiguration(output)

    s.run("chown bind:bind /etc/bind/{}".format(domain))
    s.run("/usr/sbin/named-checkzone {domain}. /etc/bind/{domain}".format(domain=domain))
    s.run("/etc/init.d/bind9 restart")
    logging.info('SUCCESSFULLY CONFIGURED {}@{}'.format(name, ip))

    s.logout()

if __name__ == '__main__':
    logging.info('Settings:\n{}'.format((dict(config.items('Global')))))
    with open('data/vps-list.txt') as f: vps_list = f.readlines()
    with open('data/email-names.txt') as f: mailbox_list = f.readlines()
    with open('data/dns_zone_template.db') as f: dns_zone_template = f.read()
    with open('data/dns_zone_template_DNSMANAGER.db') as f: dns_zone_template_DNSMANAGER = f.read()
    failed = open('var/failed.log', 'a')
    mailboxes_needed = len(vps_list) * int(config['Global']['sub_quantity']) * int(config['Global']['emails'])
    if mailboxes_needed > len(mailbox_list):
        logging.error('Not enough email names ({}). {} needed (VPS count * $sub_quantity * $). Exiting...'.format(len(mailbox_list), mailboxes_needed))
        sys.exit(1)
    # TODO: Replace with randomly generated subdomains
    for counter, line in enumerate(vps_list):
        logging.info('Configuring VPS {}/{}'.format(counter+1, len(vps_list)))
        splitted = line.strip().split(':')
        ip, domain, name, password = splitted[:4]
        if len(splitted) == 4:
            template_used = dns_zone_template
        elif len(splitted) == 5 and splitted[4] == "DNSMANAGER":
            template_used = dns_zone_template_DNSMANAGER
        else:
            logging.error('Wrong format in data/vps-list.txt (check for ":" symbols in passwords')
            sys.exit(1)

        try:
            configureServer(ip, name, password, domain, template_used)
            logging.info('DONE: {}/{}'.format(counter+1, len(vps_list)))
        except (ExceptionServerConfiguration, ExceptionServerRun,
                ExceptionServerAptGet, ExceptionSCP) as e:
            logging.error(e)
            logging.error('Error found while configuring VPS {}@{}. Skipping...'.format(name, ip))
            logging.info('FAILED: {}/{}'.format(counter+1, len(vps_list)))
            failed.write(':'.join([ip, domain, name, password]) + '\n')
            continue
